#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright 2021 Char Aznable <aznable.char.0083@gmail.com>
# Description: Utilities for pretty-printing Kokkos::UnorderedMap
#
# Distributed under terms of the 3-clause BSD license.
import re
import gdb
import argparse
import numpy as np
import pandas as pd
import pickle
from GDBKokkos.printView import (
    view2NumpyArray, getKokkosViewValueType)


def getMapValidIndices(m : gdb.Value):
    """Get the index of valid entries in a Kokkos::UnorderedMap


    Args:
        m (gdb.Value): Input Kokkos::UnorderedMap object
    Returns: tuple of np.ndarray of the valid indices, capacity of the map
    """
    bitset = view2NumpyArray(m["m_available_indexes"]["m_blocks"])
    capacity = int(m["m_available_indexes"]["m_size"])
    ans = np.where(
        np.unpackbits(bitset.view(np.uint8), bitorder="little"))[0]
    return ans, capacity


def getMapKeysVals(m : gdb.Value, depthMax : int = 3):
    """Get Kokkos::UnorderedMap's valid keys and values as numpy array


    Args:
        m (gdb.Value): Input Kokkos::UnorderedMap object
        depthMax (int): Maximal allowed recursion depth into the nested struct.
        Any struct or class below this level will be treated as byte string of
        the same size as the struct
    Returns: tuple of: the keys and values as numpy array and the capacity
    of the map
    """
    mKeys = m["m_keys"]
    mVals = m["m_values"]
    keys = view2NumpyArray(mKeys, depthMax)
    vals = view2NumpyArray(mVals, depthMax)

    ids, capacity = getMapValidIndices(m)
    keysValid = keys[ids]
    valsValid = vals[ids]
    return keysValid, valsValid, capacity

def removeByTypesFromNested(l, typesRemove : tuple):
    """Remove items in the input iteratable l which are of one of the types in
    typesRemove


    Args:
        l : Input iteratable
        typesRemove (tuple): Input types to be removed
    Returns: Iteratable with all nested items of the types removed
    """
    # remove all the bytes
    l = filter(lambda i : not isinstance(i, typesRemove), l)
    # recurse into the sublist
    l = [ removeByTypesFromNested(i, typesRemove) if isinstance(i, (list, tuple))
         else i for i in l  ]
    # clean up the empty sublists
    l = [ i for i in l if not isinstance(i, (list, tuple)) or len(i) > 0 ]
    return l


class printUnorderedMap(gdb.Command):
    """gdb command that prints a Kokkos::UnorderedMap
    """
    def __init__(self):
        # This registers the class to be used as a gdb command
        super(printUnorderedMap, self).__init__("printUnorderedMap", gdb.COMMAND_DATA)

    def parseArguments(self, inputArgs):
        parser = argparse.ArgumentParser()
        parser.add_argument("map", type=str,
                            help="Name of the UnorderedMap object")
        parser.add_argument("--noIndex", action='store_true', default=False,
                            help="Do not show the pandas.DataFrame header")
        parser.add_argument("--hideTypes", type=str, nargs='*', default=[],
                            help="When keys or values of the map have deep\
                            nested structure, GDBKokkos won't go all the way\
                            to pick up all the underlying data type to render\
                            but rather just cast anything below a certain\
                            nesting depth level, which is currently 3, to byte\
                            string. In those cases, the byte string makes\
                            rendering difficult so one can remove\
                            the byte string before printing keys or values by\
                            passing to 'bytes' to this option")
        parser.add_argument("--depthMax", type=int, default=3,
                            help="Maximal allowed recursion depth into the\
                            nested struct of view value type. Any struct or\
                            class below this level will be treated as byte \
                            string of the same size as the struct")
        return parser.parse_args(inputArgs.split())

    def invoke(self, inputArgs, from_tty):
        args = self.parseArguments(inputArgs)
        m = gdb.parse_and_eval(args.map)
        # Get the keys and vals
        keys, vals, _ = getMapKeysVals(m, args.depthMax)

        # Remove unwanted types
        typesRemove = tuple([ eval(t) for t in args.hideTypes ])
        if len(typesRemove):
            keys = removeByTypesFromNested(keys.tolist(), typesRemove)
            vals = removeByTypesFromNested(vals.tolist(), typesRemove)

        # Convert to dataframe
        # NOTE: as of Pandas version 1.2.4, Python 3.8.10 and GDB 9.2, passing
        # pd.DataFrame from python session to a gdb.Command results in garbage
        # pd.DataFrame, e.g., as if some columns are garbage collected. So we
        # have to pass the keys and vals and create the pd.DataFrame here
        mKeys = m["m_keys"]
        mVals = m["m_values"]
        TKey = getKokkosViewValueType(mKeys)
        TVal = getKokkosViewValueType(mVals)
        df = pd.DataFrame({str(TKey) : keys, str(TVal) : vals})
        with pd.option_context('display.max_seq_items', None,
                               'display.max_rows', 99999,
                               'display.max_colwidth', 2000,
                               ):
            print(df.to_string(index=not args.noIndex,
                               header=not args.noIndex))



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright 2020 Char Aznable <aznable.char.0083@gmail.com>
# Description: Utilities for pretty-printing Kokkos::View
#
# Distributed under terms of the 3-clause BSD license.
import re
import gdb
import argparse
import numpy as np
import pandas as pd
from GDBKokkos.utils import (
    name2type, foreachMemberOfClass,foreachBaseType,
    pointer2numpy)


def getKokkosViewExtent(view : gdb.Value):
    """Get the size for all dimensions of a Kokkos::View as an numpy array


    Args:
        view (gdb.Value): Input Kokkos::View
    Returns: np.ndarray of the view's extents and np.ndarray of whether a
    dimension is dynamic (1) or static (0). In case of a scalar view, the
    extents will be an empty array
    """
    # This is the ViewDimension class as template argumnet to ViewMapping
    # whose template parameters in turn indicate the size of the static view dimension
    # or zero if it's a dynamic dimension
    viewDimName = view['m_map']['m_impl_offset'].type.template_argument(0).name
    # for the dynamic dimension, get the size from ViewDimension::D0, ::D1, ...
    extents = np.array(re.findall(r'\d+', viewDimName)).astype(int)
    isDynamic = np.zeros_like(extents).astype(int)
    # There is the special case of extents being of size 0, which is a
    # specialization for ViewDimension<>, which indicates the view is a scalar
    # view (rank 0 view). In this case, we return an array of zero size
    for iDim in range(extents.size):
        if extents[iDim] > 0:
            continue
        extents[iDim] = view['m_map']['m_impl_offset']['m_dim'][f"N{iDim}"]
        isDynamic[iDim] = 1
    return extents, isDynamic


def getKokkosViewLayout(view : gdb.Value):
    """Get a Kokkos::View's memory layout as gdb.Type


    Args:
        view (gdb.Value): Input Kokkos::View
    Returns: gdb.Type of the view's layout
    """
    View = gdb.types.get_basic_type(view.type)
    return name2type(f"{View.name}::array_layout")


def getKokkosViewStrides(view : gdb.Value):
    """Get a Kokkos::View's strides as np.ndarray

    Note that Kokkos::View's m_stride member has different meaning depending on
    the View's memory layout. This function will try to return the strides in
    the same dimensionality as the View's extents unless the View defaults to no
    stride. See
    https://github.com/kokkos/kokkos/blob/master/core/src/impl/Kokkos_ViewMapping.hpp
    for details


    Args:
        view (gdb.Value): TODO

    Returns: a np.ndarray of strides, which can be zero-sized when the view is
    a scalar view

    """

    # first get the m_stride from Kokkos::View
    mStride = None
    offset = view['m_map']['m_impl_offset']
    if 'm_stride' in { m.name for m in foreachMemberOfClass(offset.type) }:
        strides = offset['m_stride']
        Tstrides = gdb.types.get_basic_type(strides.type)
        if Tstrides.code == gdb.TYPE_CODE_INT:
            mStride = np.array([strides], dtype=int)
        elif re.match(r"^Kokkos::Impl::ViewStride<\d+>$",
                      Tstrides.name) is not None:
            mStride = np.array([ strides[f"{f.name}"]
                                for f in foreachMemberOfClass(strides.type) ],
                               dtype=int)
        else:
            raise TypeError(f"Can't get strides from type {Tstrides.name}")

    extents, _ = getKokkosViewExtent(view)
    layout = getKokkosViewLayout(view)

    # if the view doesn't have stride
    if mStride is None:

        # There is nothing we can use to figure out the strides for LayoutStride
        # when we can't access m_stride
        assert layout != "Kokkos::LayoutStride",\
            f"View {str(view)} has layout stride it has no m_stride member"

        # Here, view must be LayoutLeft or LayoutRight
        # so we can safely compute the strides for all-static rank view
        aStrides = np.ones(extents.size, dtype=int)
        if layout.name == "Kokkos::LayoutRight":
            aStrides[1:] = extents[::-1][:-1]
            aStrides = np.cumprod(aStrides)[::-1]
        elif layout.name == "Kokkos::LayoutLeft":
            aStrides[1:] = extents[:-1]
            aStrides = np.cumprod(aStrides)
        else:
            raise TypeError(f"Can't handle Kokkos::View layout {layout.name}")

        return aStrides

    # then parse the strides for each dimension
    if layout.name == "Kokkos::LayoutRight":
        assert mStride.size == 1,\
            f"Can't handle other than 1 stride in layout {layout.name}"
        # mStride is the stride of rank0, which varies the slowest
        aStrides = np.ones(extents.size, dtype=int)
        aStrides[0] = mStride[0]
        aStrides[1:-1] = np.cumprod(extents[::-1][:-2])
    elif layout.name == "Kokkos::LayoutLeft":
        assert mStride.size == 1,\
            f"Can't handle other than 1 stride in layout {layout.name}"
        # mStride is the stride of rank1
        aStrides = np.ones(extents.size, dtype=int)
        aStrides[1] = mStride[0]
        aStrides[2:] = mStride[0] * np.cumprod(extents[1:-1])
    elif layout.name == "Kokkos::LayoutStride":
        assert mStride.size == extents.size,\
            f"Strides size differs from extents size in layout {layout.name}"
        aStrides = mStride
    else:
        raise TypeError(f"Can't handle Kokkos::View layout {layout.name}")
    return aStrides

def getKokkosViewTraits(view : gdb.Value):
    """Get the base Kokkos::ViewTraits of a Kokkos::View


    Args:
        view (gdb.Value): Input Kokkos::View
    Returns: gdb.Type of the view's trait type
    """
    View = gdb.types.get_basic_type(view.type)
    Trait = next(foreachBaseType(View))
    assert Trait is not None, f"Can't find traits for view of type {View.name}"
    return Trait

def getKokkosViewValueType(view : gdb.Value):
    """Get a Kokkos::View's value type


    Args:
        view (gdb.Value): Input Kokkos::View
    Returns: gdb.Type of the view's value_type
    """
    View = gdb.types.get_basic_type(view.type)
    return name2type(f"{View.name}::value_type")

def getKokkosViewSpan(view : gdb.Value):
    """Get a Kokkos::View's span

    Args:
        view (gdb.Value): Input Kokkos::View

    Returns: span of the view as int

    """
    strides = getKokkosViewStrides(view)
    extents, _ = getKokkosViewExtent(view)
    if strides.size == 0:
        # If extents.size == 0, i.e., view is a scalar view, this will return 1,
        # according to
        # https://numpy.org/doc/stable/reference/generated/numpy.prod.html
        return np.prod(extents)
    else:
        return (extents * strides).max()


def view2NumpyArray(view : gdb.Value, depthMax : int = 3):
    """Create numpy array from a Kokkos::View


    Args:
        view (gdb.Value): Input Kokkos::View
        depthMax (int): Maximal allowed recursion depth into the nested struct.
        Any struct or class below this level will be treated as byte string of
        the same size as the struct
    Returns: np.ndarray representing the view
    """
    View = gdb.types.get_basic_type(view.type)
    if re.match("Kokkos::DualView", View.name):
        # Always use device-side of view for debugging purpose
        view = view['d_view']
    elif re.match("Kokkos::View", View.name):
        pass
    else:
        raise RuntimeError(f"Don't know how to handle view of type {View.name}")

    data = view['m_map']['m_impl_handle']
    extents, _ = getKokkosViewExtent(view)
    strides = getKokkosViewStrides(view)
    dType = getKokkosViewValueType(view)
    span = getKokkosViewSpan(view)

    ans = pointer2numpy(data, dType, span, tuple(extents),
                        None if strides is None else tuple(strides),
                        depthMax)
    return ans


class printView(gdb.Command):
    """gdb command that prints a Kokkos::View
    """
    def __init__(self):
        # This registers the class to be used as a gdb command
        super(printView, self).__init__("printView", gdb.COMMAND_DATA)

    def parseArguments(self, inputArgs):
        parser = argparse.ArgumentParser()
        parser.add_argument("view", type=str,
                            help="Name of the view object")
        parser.add_argument("ranges", type=int, nargs='*', default=[],
                            help="Ranges for each of the dimension. Default to\
                            print the entire view")
        parser.add_argument("--viewTraits", action='store_true', default=False,
                            help="Print the span, extents, shape and value type\
                            of the view")
        parser.add_argument("--noIndex", action='store_true', default=False,
                            help="Do not show the rank indices when printing\
                            the view. This will render a multidimensional array\
                            into a multi-line string in a right-layout fashion")
        parser.add_argument("--depthMax", type=int, default=3,
                            help="Maximal allowed recursion depth into the\
                            nested struct of view value type. Any struct or\
                            class below this level will be treated as byte \
                            string of the same size as the struct")
        return parser.parse_args(inputArgs.split())

    def parseRanges(self, ranges : list):
        """Convert input ranges into tuple of slices of numpy array


        Args:
            ranges (list): Input list of int as slicing index
        Returns: tuple of slice objects
        """
        arr = np.array(ranges).astype(int)
        assert (not arr.size % 2) or (not arr.size % 3),\
            "Input ranges must be a series of doublets:\
            begin0, end0, begin1,end1... or triplets:\
            begin0, end0, stride0, begin1, end1, stride1..."
        if not arr.size % 2:
            arr = arr.reshape(-1, 2)
        elif not arr.size % 3:
            arr = arr.reshape(-1, 3)
        return tuple( slice( *r ) for r in arr )

    def invoke(self, inputArgs, from_tty):
        args = self.parseArguments(inputArgs)
        view = gdb.parse_and_eval(args.view)
        r = self.parseRanges(args.ranges)
        arr = view2NumpyArray(view, args.depthMax)[r]
        if args.viewTraits:
            # Print the view type traits
            extents, _ = getKokkosViewExtent(view)
            strides = getKokkosViewStrides(view)
            dType = getKokkosViewValueType(view)
            span = getKokkosViewSpan(view)
            layout = getKokkosViewLayout(view)
            print(f"# span: {span}; extents: {extents}; strides: {strides}; "
                  f"layout: {layout.name}; type: {dType}\n")
        if np.asarray(arr.shape).size <= 1:
            print(arr)
        else:
            idx = pd.MultiIndex.from_product(
                [ np.arange(dim) for dim in arr.shape[:-1] ],
                names=[f"Dim{iDim}" for iDim in range(len(arr.shape)-1)])
            if arr.dtype.names is None:
                # This is n numeric array
                df = pd.DataFrame(arr.reshape(-1, arr.shape[-1])).set_index(idx)
            else:
                # This is a structured array
                df = pd.DataFrame.from_records(
                    arr.reshape(-1, arr.shape[-1]).tolist()).set_index(idx)
            with pd.option_context('display.max_seq_items', None,
                                   'display.max_rows', 99999,
                                   'display.max_colwidth', 2000,
                                   ):
                print(df.to_string(index=not args.noIndex,
                                   header=not args.noIndex))




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright 2020 Char Aznable <aznable.char.0083@gmail.com>
# Description: setup for GDBKokkos release and install
#
# Distributed under terms of the 3-clause BSD license.
import setuptools

setuptools.setup(
    name="GDBKokkos",
    python_requires=">=3.8.10",
    version="0.1.0",
    description="GDB python modules for debugging Kokkos",
    long_description="see https://github.com/Char-Aznable/GDBKokkos",
    long_description_content_type="text/markdown",
    url="https://github.com/Char-Aznable/GDBKokkos.git",
    author="Char Aznable",
    author_email="aznable.char.0083@gmail.com",
    license="BSD 3-clause",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    packages=["GDBKokkos"],
    include_package_data=True,
    install_requires=[
        "numpy",
        "pandas",
    ],
    extras_require = {
        "tests" : ["pytest", "pytest-cov", "codecov"]
    },
    data_files = [("", ["LICENSE"])]
)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright 2020 Char Aznable <aznable.char.0083@gmail.com>
# Description: Utilities for pretty-printing Kokkos::View
#
# Distributed under terms of the 3-clause BSD license.
import gdb
from GDBKokkos.printView import(
    printView
    )
from GDBKokkos.printUnorderedMap import(
    printUnorderedMap
    )

# This registers our class to the gdb runtime at "source" time.
printView()
printUnorderedMap()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright 2020 Char Aznable <aznable.char.0083@gmail.com>
# Description: Utilities for pretty-printing Kokkos::View
#
# Distributed under terms of the 3-clause BSD license.
import numpy as np
import gdb
import re

def name2type(name : str):
    """Get gdb.Type from a string of the type name


    Args:
        name (str): Input name of the C++ type
    Returns:  gdb.Type for the requested type
    """
    symbolTuple = gdb.lookup_symbol(name)
    assert symbolTuple[0] is not None and not symbolTuple[1], f"{name} is not a type"
    t = gdb.types.get_basic_type(symbolTuple[0].type)
    return t


def foreachMemberOfClass(T : gdb.Type):
    """Generator to iterate through member of a class

    Args:
        T (gdb.Type): Input type

    Yields: field from T that is not is_base_class

    """
    for field in T.fields():
        if not field.is_base_class:
            yield field

def parseGDBArrayType(T : gdb.Type):
    """Parse an input GDB array type into the basic element type and span


    Args:
        T (gdb.Type): Input type
    Returns: tuple (gdb.Type for the element, span of the array)
    """
    assert T.code == gdb.TYPE_CODE_ARRAY, f"Type {str(T)} is not array"
    m = re.match(r'(\S+)\s*\[(\d+)\]', str(T))
    assert m is not None, f"Can't parse type {str(T)} to array"
    Element = name2type(m.group(1))
    span = int(m.group(2))
    return (Element, span)


def arithmeticType2NumpyDtype(t : gdb.Type):
    """parse basic arithmetic type into np.dtype string


    Args:
        t (gdb.Type): Input type
    Returns: a tuple (element type, span) where span can be the extent of the
    input array type or None if input is scalar type
    """

    T = gdb.types.get_basic_type(t)

    # parse array
    span = None
    if T.code == gdb.TYPE_CODE_ARRAY:
        T, span = parseGDBArrayType(T)

    sizeT = T.sizeof
    isSigned = re.search(r"unsigned", T.name) is None if T.code == gdb.TYPE_CODE_INT else None

    typeCode2TypeStr = {
        gdb.TYPE_CODE_INT : f"|{'i' if isSigned else 'u'}{sizeT}",
        gdb.TYPE_CODE_FLT : f"|f{sizeT}",
        gdb.TYPE_CODE_BOOL : f"|b{sizeT}",
        }

    if not T.code in typeCode2TypeStr:
        # This is likely some other types that can't be rendered into numpy
        # array so we just render it into string. See
        # https://github.com/bminor/binutils-gdb/blob/777b054cf93ad2525b891ea15bbf8d5cd6a56339/gdb/gdbtypes.h#L95
        # for the list of gdb type code
        return (f"|S{sizeT}", span)

    return (typeCode2TypeStr[T.code], span)


def struct2dtype(T : gdb.Type, depth : int = 0, depthMax : int = 3):
    """Convert a gdb.Type to a list of np.dtype with potentially nested struct

    Nested struct are parsed as what would appear in a numpy structured array
    with the output name prefixed with outter class scope name

    Args:
        T (gdb.Type): Input type to be parsed
        depth (int): Current recursion depth of calling this function
        depthMax (int): Maximal allowed recursion depth into the nested struct.
        Any struct or class below this level will be treated as byte string of
        the same size as the struct
    Returns: list [name, np.dtype]
    """
    assert T.code == gdb.TYPE_CODE_STRUCT, f"Input type {str(T)} is not struct"

    ans = []
    for f in foreachMemberOfClass(T):
        name = f.name
        t = gdb.types.get_basic_type(f.type)
        if t.code == gdb.TYPE_CODE_STRUCT and depth < depthMax:
            ans.append((name, struct2dtype(t, depth + 1, depthMax)))
        else:
            dtype, span = arithmeticType2NumpyDtype(t)
            ans.append((name, dtype) if span is None else (name, dtype, span))
    return ans


def type2dtype(T : gdb.Type, depthMax : int = 3):
    """Convert input gdb.Type to np.dtype

    For simple scalar type, this will return a single string of np.dtype

    For C array of struct (potentially nested), it will output a list of
    np.dtype fields with the name of each field being the struct member variable
    (potentially prefixed with outter class name if nesting). Array type will be
    suffixed with the span of the array

    Args:
        T (gdb.Type): Input type
        depthMax (int): Maximal allowed recursion depth into the nested struct.
        Any struct or class below this level will be treated as byte string of
        the same size as the struct
    Returns: str or list of np.dtype
    """

    #TODO: convert structed array to pandas dataframe:
    # pd.DataFrame.from_records(x.tolist(), columns=x.dtype.names)

    dtypes = None
    if T.code == gdb.TYPE_CODE_STRUCT:
        # this is a struct, potentially nested. we first flatten out the nested
        # structs and create a numpy structured array using the fields in the
        # flatten struct
        dtypes = struct2dtype(T, 0, depthMax)
    else:
        dtype, span = arithmeticType2NumpyDtype(T)
        if span is None:
            # this is a scalar type so we use dtypes to create a normal numpy
            # array
            dtypes = dtype
        else:
            # this is an array type so we use dtypes to create a numpy
            # structured array
            dtypes = [(T.name, dtype, span)]
    return dtypes


def foreachBaseType(T : gdb.Type):
    """Generator to iterate through the base types of a type

    Args:
        T (gdb.Type): Input type

    Yields: field from T that is is_base_class

    """
    for field in T.fields():
        if field.is_base_class:
            yield field.type


def pointer2numpy(addr : int, t : gdb.Type, span : int, shape : tuple,
                  strides : tuple = None, depthMax : int = 3):
    """Create a numpy array from given memory address, type and span

    Args:
        addr (int): The address in the C program's memory space
        t (gdb.Type): The C-type of the array element
        span (int): The total span of the array in the unit of number of
        elements.
        shape (tuple): The shape of the output array
        strides (tuple) : strides of each dimension of the array in the unit of
        number of elements
        depthMax (int): Maximal allowed recursion depth into the nested struct.
        Any struct or class below this level will be treated as byte string of
        the same size as the struct

    Returns: np.ndarray representing the given memory chunk

    """

    # Get the underlying type, removing typdefs
    T = gdb.types.get_basic_type(t)

    sizeT = T.sizeof
    sizeTotal = span * sizeT

    typestr = type2dtype(T, depthMax)
    dtype = np.dtype(typestr, align=True)

    # in case the alignment assumption failed, e.g., due to the compiler, check
    # if the size of the type is consistent
    assert dtype.itemsize == sizeT,\
        f"np.dtype reports type {str(T)} has size {dtype.itemsize} but its "\
        f"actual size is {sizeT}. It's likely that the alignment assumption "\
        f"of np.dtype failed."

    # create a gdb.Value representing the array, i.e., arr.type == T[span]
    # arr = gdb.parse_and_eval(f"*(({T.name} *){addr})@{span}")
    # TODO: to support CUDA global memory space array, use something like:
    # arr = gdb.parse_and_eval(f"*((@global {T.name} *){addr})@{span}")
    # or create the buffer with
    # buffer = gdb.selected_inferior().read_memory(f"(@global {T.name} *){addr}", sizeTotal)
    buffer = gdb.selected_inferior().read_memory(addr, sizeTotal)
    # copy the array over
    # This is equivlant to 
    # aTmp = np.array([arr[i] for i in range(span)], dtype=dtype)
    # for simple dtype. For more complex type with nested structure, the
    # element-wise copying would involve nested loops and is less favored to
    # direct byte copying here
    aTmp = np.frombuffer(buffer, dtype=dtype).copy()
    ans = np.lib.stride_tricks.as_strided(
        aTmp,
        shape=shape, strides=np.asarray(strides) * sizeT,
        writeable=False)

    return ans
